/*-------------------------------------------------------------------------
 *
 * queryjumblefuncs.switch.c
 *    Generated node infrastructure code
 *
 * Portions Copyright (c) 1996-2025, PostgreSQL Global Development Group
 * Portions Copyright (c) 1994, Regents of the University of California
 *
 * NOTES
 *  ******************************
 *  *** DO NOT EDIT THIS FILE! ***
 *  ******************************
 *
 *  It has been GENERATED by src/backend/nodes/gen_node_support.pl
 *
 *-------------------------------------------------------------------------
 */
			case T_Alias:
				_jumbleAlias(jstate, expr);
				break;
			case T_RangeVar:
				_jumbleRangeVar(jstate, expr);
				break;
			case T_TableFunc:
				_jumbleTableFunc(jstate, expr);
				break;
			case T_IntoClause:
				_jumbleIntoClause(jstate, expr);
				break;
			case T_Var:
				_jumbleVar(jstate, expr);
				break;
			case T_Const:
				_jumbleConst(jstate, expr);
				break;
			case T_Param:
				_jumbleParam(jstate, expr);
				break;
			case T_Aggref:
				_jumbleAggref(jstate, expr);
				break;
			case T_GroupingFunc:
				_jumbleGroupingFunc(jstate, expr);
				break;
			case T_WindowFunc:
				_jumbleWindowFunc(jstate, expr);
				break;
			case T_WindowFuncRunCondition:
				_jumbleWindowFuncRunCondition(jstate, expr);
				break;
			case T_MergeSupportFunc:
				_jumbleMergeSupportFunc(jstate, expr);
				break;
			case T_SubscriptingRef:
				_jumbleSubscriptingRef(jstate, expr);
				break;
			case T_FuncExpr:
				_jumbleFuncExpr(jstate, expr);
				break;
			case T_NamedArgExpr:
				_jumbleNamedArgExpr(jstate, expr);
				break;
			case T_OpExpr:
				_jumbleOpExpr(jstate, expr);
				break;
			case T_DistinctExpr:
				_jumbleDistinctExpr(jstate, expr);
				break;
			case T_NullIfExpr:
				_jumbleNullIfExpr(jstate, expr);
				break;
			case T_ScalarArrayOpExpr:
				_jumbleScalarArrayOpExpr(jstate, expr);
				break;
			case T_BoolExpr:
				_jumbleBoolExpr(jstate, expr);
				break;
			case T_SubLink:
				_jumbleSubLink(jstate, expr);
				break;
			case T_FieldSelect:
				_jumbleFieldSelect(jstate, expr);
				break;
			case T_FieldStore:
				_jumbleFieldStore(jstate, expr);
				break;
			case T_RelabelType:
				_jumbleRelabelType(jstate, expr);
				break;
			case T_CoerceViaIO:
				_jumbleCoerceViaIO(jstate, expr);
				break;
			case T_ArrayCoerceExpr:
				_jumbleArrayCoerceExpr(jstate, expr);
				break;
			case T_ConvertRowtypeExpr:
				_jumbleConvertRowtypeExpr(jstate, expr);
				break;
			case T_CollateExpr:
				_jumbleCollateExpr(jstate, expr);
				break;
			case T_CaseExpr:
				_jumbleCaseExpr(jstate, expr);
				break;
			case T_CaseWhen:
				_jumbleCaseWhen(jstate, expr);
				break;
			case T_CaseTestExpr:
				_jumbleCaseTestExpr(jstate, expr);
				break;
			case T_ArrayExpr:
				_jumbleArrayExpr(jstate, expr);
				break;
			case T_RowExpr:
				_jumbleRowExpr(jstate, expr);
				break;
			case T_RowCompareExpr:
				_jumbleRowCompareExpr(jstate, expr);
				break;
			case T_CoalesceExpr:
				_jumbleCoalesceExpr(jstate, expr);
				break;
			case T_MinMaxExpr:
				_jumbleMinMaxExpr(jstate, expr);
				break;
			case T_SQLValueFunction:
				_jumbleSQLValueFunction(jstate, expr);
				break;
			case T_XmlExpr:
				_jumbleXmlExpr(jstate, expr);
				break;
			case T_JsonFormat:
				_jumbleJsonFormat(jstate, expr);
				break;
			case T_JsonReturning:
				_jumbleJsonReturning(jstate, expr);
				break;
			case T_JsonValueExpr:
				_jumbleJsonValueExpr(jstate, expr);
				break;
			case T_JsonConstructorExpr:
				_jumbleJsonConstructorExpr(jstate, expr);
				break;
			case T_JsonIsPredicate:
				_jumbleJsonIsPredicate(jstate, expr);
				break;
			case T_JsonBehavior:
				_jumbleJsonBehavior(jstate, expr);
				break;
			case T_JsonExpr:
				_jumbleJsonExpr(jstate, expr);
				break;
			case T_JsonTablePath:
				_jumbleJsonTablePath(jstate, expr);
				break;
			case T_JsonTablePathScan:
				_jumbleJsonTablePathScan(jstate, expr);
				break;
			case T_JsonTableSiblingJoin:
				_jumbleJsonTableSiblingJoin(jstate, expr);
				break;
			case T_NullTest:
				_jumbleNullTest(jstate, expr);
				break;
			case T_BooleanTest:
				_jumbleBooleanTest(jstate, expr);
				break;
			case T_MergeAction:
				_jumbleMergeAction(jstate, expr);
				break;
			case T_CoerceToDomain:
				_jumbleCoerceToDomain(jstate, expr);
				break;
			case T_CoerceToDomainValue:
				_jumbleCoerceToDomainValue(jstate, expr);
				break;
			case T_SetToDefault:
				_jumbleSetToDefault(jstate, expr);
				break;
			case T_CurrentOfExpr:
				_jumbleCurrentOfExpr(jstate, expr);
				break;
			case T_NextValueExpr:
				_jumbleNextValueExpr(jstate, expr);
				break;
			case T_InferenceElem:
				_jumbleInferenceElem(jstate, expr);
				break;
			case T_ReturningExpr:
				_jumbleReturningExpr(jstate, expr);
				break;
			case T_TargetEntry:
				_jumbleTargetEntry(jstate, expr);
				break;
			case T_RangeTblRef:
				_jumbleRangeTblRef(jstate, expr);
				break;
			case T_JoinExpr:
				_jumbleJoinExpr(jstate, expr);
				break;
			case T_FromExpr:
				_jumbleFromExpr(jstate, expr);
				break;
			case T_OnConflictExpr:
				_jumbleOnConflictExpr(jstate, expr);
				break;
			case T_Query:
				_jumbleQuery(jstate, expr);
				break;
			case T_TypeName:
				_jumbleTypeName(jstate, expr);
				break;
			case T_ColumnRef:
				_jumbleColumnRef(jstate, expr);
				break;
			case T_ParamRef:
				_jumbleParamRef(jstate, expr);
				break;
			case T_A_Expr:
				_jumbleA_Expr(jstate, expr);
				break;
			case T_A_Const:
				_jumbleA_Const(jstate, expr);
				break;
			case T_TypeCast:
				_jumbleTypeCast(jstate, expr);
				break;
			case T_CollateClause:
				_jumbleCollateClause(jstate, expr);
				break;
			case T_RoleSpec:
				_jumbleRoleSpec(jstate, expr);
				break;
			case T_FuncCall:
				_jumbleFuncCall(jstate, expr);
				break;
			case T_A_Star:
				_jumbleA_Star(jstate, expr);
				break;
			case T_A_Indices:
				_jumbleA_Indices(jstate, expr);
				break;
			case T_A_Indirection:
				_jumbleA_Indirection(jstate, expr);
				break;
			case T_A_ArrayExpr:
				_jumbleA_ArrayExpr(jstate, expr);
				break;
			case T_ResTarget:
				_jumbleResTarget(jstate, expr);
				break;
			case T_MultiAssignRef:
				_jumbleMultiAssignRef(jstate, expr);
				break;
			case T_SortBy:
				_jumbleSortBy(jstate, expr);
				break;
			case T_WindowDef:
				_jumbleWindowDef(jstate, expr);
				break;
			case T_RangeSubselect:
				_jumbleRangeSubselect(jstate, expr);
				break;
			case T_RangeFunction:
				_jumbleRangeFunction(jstate, expr);
				break;
			case T_RangeTableFunc:
				_jumbleRangeTableFunc(jstate, expr);
				break;
			case T_RangeTableFuncCol:
				_jumbleRangeTableFuncCol(jstate, expr);
				break;
			case T_RangeTableSample:
				_jumbleRangeTableSample(jstate, expr);
				break;
			case T_ColumnDef:
				_jumbleColumnDef(jstate, expr);
				break;
			case T_TableLikeClause:
				_jumbleTableLikeClause(jstate, expr);
				break;
			case T_IndexElem:
				_jumbleIndexElem(jstate, expr);
				break;
			case T_DefElem:
				_jumbleDefElem(jstate, expr);
				break;
			case T_LockingClause:
				_jumbleLockingClause(jstate, expr);
				break;
			case T_XmlSerialize:
				_jumbleXmlSerialize(jstate, expr);
				break;
			case T_PartitionElem:
				_jumblePartitionElem(jstate, expr);
				break;
			case T_PartitionSpec:
				_jumblePartitionSpec(jstate, expr);
				break;
			case T_PartitionBoundSpec:
				_jumblePartitionBoundSpec(jstate, expr);
				break;
			case T_PartitionRangeDatum:
				_jumblePartitionRangeDatum(jstate, expr);
				break;
			case T_PartitionCmd:
				_jumblePartitionCmd(jstate, expr);
				break;
			case T_RangeTblEntry:
				_jumbleRangeTblEntry(jstate, expr);
				break;
			case T_RTEPermissionInfo:
				_jumbleRTEPermissionInfo(jstate, expr);
				break;
			case T_RangeTblFunction:
				_jumbleRangeTblFunction(jstate, expr);
				break;
			case T_TableSampleClause:
				_jumbleTableSampleClause(jstate, expr);
				break;
			case T_WithCheckOption:
				_jumbleWithCheckOption(jstate, expr);
				break;
			case T_SortGroupClause:
				_jumbleSortGroupClause(jstate, expr);
				break;
			case T_GroupingSet:
				_jumbleGroupingSet(jstate, expr);
				break;
			case T_WindowClause:
				_jumbleWindowClause(jstate, expr);
				break;
			case T_RowMarkClause:
				_jumbleRowMarkClause(jstate, expr);
				break;
			case T_WithClause:
				_jumbleWithClause(jstate, expr);
				break;
			case T_InferClause:
				_jumbleInferClause(jstate, expr);
				break;
			case T_OnConflictClause:
				_jumbleOnConflictClause(jstate, expr);
				break;
			case T_CTESearchClause:
				_jumbleCTESearchClause(jstate, expr);
				break;
			case T_CTECycleClause:
				_jumbleCTECycleClause(jstate, expr);
				break;
			case T_CommonTableExpr:
				_jumbleCommonTableExpr(jstate, expr);
				break;
			case T_MergeWhenClause:
				_jumbleMergeWhenClause(jstate, expr);
				break;
			case T_ReturningOption:
				_jumbleReturningOption(jstate, expr);
				break;
			case T_ReturningClause:
				_jumbleReturningClause(jstate, expr);
				break;
			case T_TriggerTransition:
				_jumbleTriggerTransition(jstate, expr);
				break;
			case T_JsonOutput:
				_jumbleJsonOutput(jstate, expr);
				break;
			case T_JsonArgument:
				_jumbleJsonArgument(jstate, expr);
				break;
			case T_JsonFuncExpr:
				_jumbleJsonFuncExpr(jstate, expr);
				break;
			case T_JsonTablePathSpec:
				_jumbleJsonTablePathSpec(jstate, expr);
				break;
			case T_JsonTable:
				_jumbleJsonTable(jstate, expr);
				break;
			case T_JsonTableColumn:
				_jumbleJsonTableColumn(jstate, expr);
				break;
			case T_JsonKeyValue:
				_jumbleJsonKeyValue(jstate, expr);
				break;
			case T_JsonParseExpr:
				_jumbleJsonParseExpr(jstate, expr);
				break;
			case T_JsonScalarExpr:
				_jumbleJsonScalarExpr(jstate, expr);
				break;
			case T_JsonSerializeExpr:
				_jumbleJsonSerializeExpr(jstate, expr);
				break;
			case T_JsonObjectConstructor:
				_jumbleJsonObjectConstructor(jstate, expr);
				break;
			case T_JsonArrayConstructor:
				_jumbleJsonArrayConstructor(jstate, expr);
				break;
			case T_JsonArrayQueryConstructor:
				_jumbleJsonArrayQueryConstructor(jstate, expr);
				break;
			case T_JsonAggConstructor:
				_jumbleJsonAggConstructor(jstate, expr);
				break;
			case T_JsonObjectAgg:
				_jumbleJsonObjectAgg(jstate, expr);
				break;
			case T_JsonArrayAgg:
				_jumbleJsonArrayAgg(jstate, expr);
				break;
			case T_InsertStmt:
				_jumbleInsertStmt(jstate, expr);
				break;
			case T_DeleteStmt:
				_jumbleDeleteStmt(jstate, expr);
				break;
			case T_UpdateStmt:
				_jumbleUpdateStmt(jstate, expr);
				break;
			case T_MergeStmt:
				_jumbleMergeStmt(jstate, expr);
				break;
			case T_SelectStmt:
				_jumbleSelectStmt(jstate, expr);
				break;
			case T_SetOperationStmt:
				_jumbleSetOperationStmt(jstate, expr);
				break;
			case T_ReturnStmt:
				_jumbleReturnStmt(jstate, expr);
				break;
			case T_PLAssignStmt:
				_jumblePLAssignStmt(jstate, expr);
				break;
			case T_CreateSchemaStmt:
				_jumbleCreateSchemaStmt(jstate, expr);
				break;
			case T_AlterTableStmt:
				_jumbleAlterTableStmt(jstate, expr);
				break;
			case T_AlterTableCmd:
				_jumbleAlterTableCmd(jstate, expr);
				break;
			case T_ATAlterConstraint:
				_jumbleATAlterConstraint(jstate, expr);
				break;
			case T_ReplicaIdentityStmt:
				_jumbleReplicaIdentityStmt(jstate, expr);
				break;
			case T_AlterCollationStmt:
				_jumbleAlterCollationStmt(jstate, expr);
				break;
			case T_AlterDomainStmt:
				_jumbleAlterDomainStmt(jstate, expr);
				break;
			case T_GrantStmt:
				_jumbleGrantStmt(jstate, expr);
				break;
			case T_ObjectWithArgs:
				_jumbleObjectWithArgs(jstate, expr);
				break;
			case T_AccessPriv:
				_jumbleAccessPriv(jstate, expr);
				break;
			case T_GrantRoleStmt:
				_jumbleGrantRoleStmt(jstate, expr);
				break;
			case T_AlterDefaultPrivilegesStmt:
				_jumbleAlterDefaultPrivilegesStmt(jstate, expr);
				break;
			case T_CopyStmt:
				_jumbleCopyStmt(jstate, expr);
				break;
			case T_VariableSetStmt:
				_jumbleVariableSetStmt(jstate, expr);
				break;
			case T_VariableShowStmt:
				_jumbleVariableShowStmt(jstate, expr);
				break;
			case T_CreateStmt:
				_jumbleCreateStmt(jstate, expr);
				break;
			case T_Constraint:
				_jumbleConstraint(jstate, expr);
				break;
			case T_CreateTableSpaceStmt:
				_jumbleCreateTableSpaceStmt(jstate, expr);
				break;
			case T_DropTableSpaceStmt:
				_jumbleDropTableSpaceStmt(jstate, expr);
				break;
			case T_AlterTableSpaceOptionsStmt:
				_jumbleAlterTableSpaceOptionsStmt(jstate, expr);
				break;
			case T_AlterTableMoveAllStmt:
				_jumbleAlterTableMoveAllStmt(jstate, expr);
				break;
			case T_CreateExtensionStmt:
				_jumbleCreateExtensionStmt(jstate, expr);
				break;
			case T_AlterExtensionStmt:
				_jumbleAlterExtensionStmt(jstate, expr);
				break;
			case T_AlterExtensionContentsStmt:
				_jumbleAlterExtensionContentsStmt(jstate, expr);
				break;
			case T_CreateFdwStmt:
				_jumbleCreateFdwStmt(jstate, expr);
				break;
			case T_AlterFdwStmt:
				_jumbleAlterFdwStmt(jstate, expr);
				break;
			case T_CreateForeignServerStmt:
				_jumbleCreateForeignServerStmt(jstate, expr);
				break;
			case T_AlterForeignServerStmt:
				_jumbleAlterForeignServerStmt(jstate, expr);
				break;
			case T_CreateForeignTableStmt:
				_jumbleCreateForeignTableStmt(jstate, expr);
				break;
			case T_CreateUserMappingStmt:
				_jumbleCreateUserMappingStmt(jstate, expr);
				break;
			case T_AlterUserMappingStmt:
				_jumbleAlterUserMappingStmt(jstate, expr);
				break;
			case T_DropUserMappingStmt:
				_jumbleDropUserMappingStmt(jstate, expr);
				break;
			case T_ImportForeignSchemaStmt:
				_jumbleImportForeignSchemaStmt(jstate, expr);
				break;
			case T_CreatePolicyStmt:
				_jumbleCreatePolicyStmt(jstate, expr);
				break;
			case T_AlterPolicyStmt:
				_jumbleAlterPolicyStmt(jstate, expr);
				break;
			case T_CreateAmStmt:
				_jumbleCreateAmStmt(jstate, expr);
				break;
			case T_CreateTrigStmt:
				_jumbleCreateTrigStmt(jstate, expr);
				break;
			case T_CreateEventTrigStmt:
				_jumbleCreateEventTrigStmt(jstate, expr);
				break;
			case T_AlterEventTrigStmt:
				_jumbleAlterEventTrigStmt(jstate, expr);
				break;
			case T_CreatePLangStmt:
				_jumbleCreatePLangStmt(jstate, expr);
				break;
			case T_CreateRoleStmt:
				_jumbleCreateRoleStmt(jstate, expr);
				break;
			case T_AlterRoleStmt:
				_jumbleAlterRoleStmt(jstate, expr);
				break;
			case T_AlterRoleSetStmt:
				_jumbleAlterRoleSetStmt(jstate, expr);
				break;
			case T_DropRoleStmt:
				_jumbleDropRoleStmt(jstate, expr);
				break;
			case T_CreateSeqStmt:
				_jumbleCreateSeqStmt(jstate, expr);
				break;
			case T_AlterSeqStmt:
				_jumbleAlterSeqStmt(jstate, expr);
				break;
			case T_DefineStmt:
				_jumbleDefineStmt(jstate, expr);
				break;
			case T_CreateDomainStmt:
				_jumbleCreateDomainStmt(jstate, expr);
				break;
			case T_CreateOpClassStmt:
				_jumbleCreateOpClassStmt(jstate, expr);
				break;
			case T_CreateOpClassItem:
				_jumbleCreateOpClassItem(jstate, expr);
				break;
			case T_CreateOpFamilyStmt:
				_jumbleCreateOpFamilyStmt(jstate, expr);
				break;
			case T_AlterOpFamilyStmt:
				_jumbleAlterOpFamilyStmt(jstate, expr);
				break;
			case T_DropStmt:
				_jumbleDropStmt(jstate, expr);
				break;
			case T_TruncateStmt:
				_jumbleTruncateStmt(jstate, expr);
				break;
			case T_CommentStmt:
				_jumbleCommentStmt(jstate, expr);
				break;
			case T_SecLabelStmt:
				_jumbleSecLabelStmt(jstate, expr);
				break;
			case T_DeclareCursorStmt:
				_jumbleDeclareCursorStmt(jstate, expr);
				break;
			case T_ClosePortalStmt:
				_jumbleClosePortalStmt(jstate, expr);
				break;
			case T_FetchStmt:
				_jumbleFetchStmt(jstate, expr);
				break;
			case T_IndexStmt:
				_jumbleIndexStmt(jstate, expr);
				break;
			case T_CreateStatsStmt:
				_jumbleCreateStatsStmt(jstate, expr);
				break;
			case T_StatsElem:
				_jumbleStatsElem(jstate, expr);
				break;
			case T_AlterStatsStmt:
				_jumbleAlterStatsStmt(jstate, expr);
				break;
			case T_CreateFunctionStmt:
				_jumbleCreateFunctionStmt(jstate, expr);
				break;
			case T_FunctionParameter:
				_jumbleFunctionParameter(jstate, expr);
				break;
			case T_AlterFunctionStmt:
				_jumbleAlterFunctionStmt(jstate, expr);
				break;
			case T_DoStmt:
				_jumbleDoStmt(jstate, expr);
				break;
			case T_CallStmt:
				_jumbleCallStmt(jstate, expr);
				break;
			case T_RenameStmt:
				_jumbleRenameStmt(jstate, expr);
				break;
			case T_AlterObjectDependsStmt:
				_jumbleAlterObjectDependsStmt(jstate, expr);
				break;
			case T_AlterObjectSchemaStmt:
				_jumbleAlterObjectSchemaStmt(jstate, expr);
				break;
			case T_AlterOwnerStmt:
				_jumbleAlterOwnerStmt(jstate, expr);
				break;
			case T_AlterOperatorStmt:
				_jumbleAlterOperatorStmt(jstate, expr);
				break;
			case T_AlterTypeStmt:
				_jumbleAlterTypeStmt(jstate, expr);
				break;
			case T_RuleStmt:
				_jumbleRuleStmt(jstate, expr);
				break;
			case T_NotifyStmt:
				_jumbleNotifyStmt(jstate, expr);
				break;
			case T_ListenStmt:
				_jumbleListenStmt(jstate, expr);
				break;
			case T_UnlistenStmt:
				_jumbleUnlistenStmt(jstate, expr);
				break;
			case T_TransactionStmt:
				_jumbleTransactionStmt(jstate, expr);
				break;
			case T_CompositeTypeStmt:
				_jumbleCompositeTypeStmt(jstate, expr);
				break;
			case T_CreateEnumStmt:
				_jumbleCreateEnumStmt(jstate, expr);
				break;
			case T_CreateRangeStmt:
				_jumbleCreateRangeStmt(jstate, expr);
				break;
			case T_AlterEnumStmt:
				_jumbleAlterEnumStmt(jstate, expr);
				break;
			case T_ViewStmt:
				_jumbleViewStmt(jstate, expr);
				break;
			case T_LoadStmt:
				_jumbleLoadStmt(jstate, expr);
				break;
			case T_CreatedbStmt:
				_jumbleCreatedbStmt(jstate, expr);
				break;
			case T_AlterDatabaseStmt:
				_jumbleAlterDatabaseStmt(jstate, expr);
				break;
			case T_AlterDatabaseRefreshCollStmt:
				_jumbleAlterDatabaseRefreshCollStmt(jstate, expr);
				break;
			case T_AlterDatabaseSetStmt:
				_jumbleAlterDatabaseSetStmt(jstate, expr);
				break;
			case T_DropdbStmt:
				_jumbleDropdbStmt(jstate, expr);
				break;
			case T_AlterSystemStmt:
				_jumbleAlterSystemStmt(jstate, expr);
				break;
			case T_ClusterStmt:
				_jumbleClusterStmt(jstate, expr);
				break;
			case T_VacuumStmt:
				_jumbleVacuumStmt(jstate, expr);
				break;
			case T_VacuumRelation:
				_jumbleVacuumRelation(jstate, expr);
				break;
			case T_ExplainStmt:
				_jumbleExplainStmt(jstate, expr);
				break;
			case T_CreateTableAsStmt:
				_jumbleCreateTableAsStmt(jstate, expr);
				break;
			case T_RefreshMatViewStmt:
				_jumbleRefreshMatViewStmt(jstate, expr);
				break;
			case T_CheckPointStmt:
				_jumbleCheckPointStmt(jstate, expr);
				break;
			case T_DiscardStmt:
				_jumbleDiscardStmt(jstate, expr);
				break;
			case T_LockStmt:
				_jumbleLockStmt(jstate, expr);
				break;
			case T_ConstraintsSetStmt:
				_jumbleConstraintsSetStmt(jstate, expr);
				break;
			case T_ReindexStmt:
				_jumbleReindexStmt(jstate, expr);
				break;
			case T_CreateConversionStmt:
				_jumbleCreateConversionStmt(jstate, expr);
				break;
			case T_CreateCastStmt:
				_jumbleCreateCastStmt(jstate, expr);
				break;
			case T_CreateTransformStmt:
				_jumbleCreateTransformStmt(jstate, expr);
				break;
			case T_PrepareStmt:
				_jumblePrepareStmt(jstate, expr);
				break;
			case T_ExecuteStmt:
				_jumbleExecuteStmt(jstate, expr);
				break;
			case T_DeallocateStmt:
				_jumbleDeallocateStmt(jstate, expr);
				break;
			case T_DropOwnedStmt:
				_jumbleDropOwnedStmt(jstate, expr);
				break;
			case T_ReassignOwnedStmt:
				_jumbleReassignOwnedStmt(jstate, expr);
				break;
			case T_AlterTSDictionaryStmt:
				_jumbleAlterTSDictionaryStmt(jstate, expr);
				break;
			case T_AlterTSConfigurationStmt:
				_jumbleAlterTSConfigurationStmt(jstate, expr);
				break;
			case T_PublicationTable:
				_jumblePublicationTable(jstate, expr);
				break;
			case T_PublicationObjSpec:
				_jumblePublicationObjSpec(jstate, expr);
				break;
			case T_CreatePublicationStmt:
				_jumbleCreatePublicationStmt(jstate, expr);
				break;
			case T_AlterPublicationStmt:
				_jumbleAlterPublicationStmt(jstate, expr);
				break;
			case T_CreateSubscriptionStmt:
				_jumbleCreateSubscriptionStmt(jstate, expr);
				break;
			case T_AlterSubscriptionStmt:
				_jumbleAlterSubscriptionStmt(jstate, expr);
				break;
			case T_DropSubscriptionStmt:
				_jumbleDropSubscriptionStmt(jstate, expr);
				break;
			case T_GroupByOrdering:
				_jumbleGroupByOrdering(jstate, expr);
				break;
			case T_ExtensibleNode:
				_jumbleExtensibleNode(jstate, expr);
				break;
			case T_Integer:
				_jumbleInteger(jstate, expr);
				break;
			case T_Float:
				_jumbleFloat(jstate, expr);
				break;
			case T_Boolean:
				_jumbleBoolean(jstate, expr);
				break;
			case T_String:
				_jumbleString(jstate, expr);
				break;
			case T_BitString:
				_jumbleBitString(jstate, expr);
				break;
