/*-------------------------------------------------------------------------
 *
 * readfuncs.switch.c
 *    Generated node infrastructure code
 *
 * Portions Copyright (c) 1996-2025, PostgreSQL Global Development Group
 * Portions Copyright (c) 1994, Regents of the University of California
 *
 * NOTES
 *  ******************************
 *  *** DO NOT EDIT THIS FILE! ***
 *  ******************************
 *
 *  It has been GENERATED by src/backend/nodes/gen_node_support.pl
 *
 *-------------------------------------------------------------------------
 */
	if (MATCH("ALIAS", 5))
		return (Node *) _readAlias();
	if (MATCH("RANGEVAR", 8))
		return (Node *) _readRangeVar();
	if (MATCH("TABLEFUNC", 9))
		return (Node *) _readTableFunc();
	if (MATCH("INTOCLAUSE", 10))
		return (Node *) _readIntoClause();
	if (MATCH("VAR", 3))
		return (Node *) _readVar();
	if (MATCH("CONST", 5))
		return (Node *) _readConst();
	if (MATCH("PARAM", 5))
		return (Node *) _readParam();
	if (MATCH("AGGREF", 6))
		return (Node *) _readAggref();
	if (MATCH("GROUPINGFUNC", 12))
		return (Node *) _readGroupingFunc();
	if (MATCH("WINDOWFUNC", 10))
		return (Node *) _readWindowFunc();
	if (MATCH("WINDOWFUNCRUNCONDITION", 22))
		return (Node *) _readWindowFuncRunCondition();
	if (MATCH("MERGESUPPORTFUNC", 16))
		return (Node *) _readMergeSupportFunc();
	if (MATCH("SUBSCRIPTINGREF", 15))
		return (Node *) _readSubscriptingRef();
	if (MATCH("FUNCEXPR", 8))
		return (Node *) _readFuncExpr();
	if (MATCH("NAMEDARGEXPR", 12))
		return (Node *) _readNamedArgExpr();
	if (MATCH("OPEXPR", 6))
		return (Node *) _readOpExpr();
	if (MATCH("DISTINCTEXPR", 12))
		return (Node *) _readDistinctExpr();
	if (MATCH("NULLIFEXPR", 10))
		return (Node *) _readNullIfExpr();
	if (MATCH("SCALARARRAYOPEXPR", 17))
		return (Node *) _readScalarArrayOpExpr();
	if (MATCH("BOOLEXPR", 8))
		return (Node *) _readBoolExpr();
	if (MATCH("SUBLINK", 7))
		return (Node *) _readSubLink();
	if (MATCH("SUBPLAN", 7))
		return (Node *) _readSubPlan();
	if (MATCH("ALTERNATIVESUBPLAN", 18))
		return (Node *) _readAlternativeSubPlan();
	if (MATCH("FIELDSELECT", 11))
		return (Node *) _readFieldSelect();
	if (MATCH("FIELDSTORE", 10))
		return (Node *) _readFieldStore();
	if (MATCH("RELABELTYPE", 11))
		return (Node *) _readRelabelType();
	if (MATCH("COERCEVIAIO", 11))
		return (Node *) _readCoerceViaIO();
	if (MATCH("ARRAYCOERCEEXPR", 15))
		return (Node *) _readArrayCoerceExpr();
	if (MATCH("CONVERTROWTYPEEXPR", 18))
		return (Node *) _readConvertRowtypeExpr();
	if (MATCH("COLLATEEXPR", 11))
		return (Node *) _readCollateExpr();
	if (MATCH("CASEEXPR", 8))
		return (Node *) _readCaseExpr();
	if (MATCH("CASEWHEN", 8))
		return (Node *) _readCaseWhen();
	if (MATCH("CASETESTEXPR", 12))
		return (Node *) _readCaseTestExpr();
	if (MATCH("ARRAYEXPR", 9))
		return (Node *) _readArrayExpr();
	if (MATCH("ROWEXPR", 7))
		return (Node *) _readRowExpr();
	if (MATCH("ROWCOMPAREEXPR", 14))
		return (Node *) _readRowCompareExpr();
	if (MATCH("COALESCEEXPR", 12))
		return (Node *) _readCoalesceExpr();
	if (MATCH("MINMAXEXPR", 10))
		return (Node *) _readMinMaxExpr();
	if (MATCH("SQLVALUEFUNCTION", 16))
		return (Node *) _readSQLValueFunction();
	if (MATCH("XMLEXPR", 7))
		return (Node *) _readXmlExpr();
	if (MATCH("JSONFORMAT", 10))
		return (Node *) _readJsonFormat();
	if (MATCH("JSONRETURNING", 13))
		return (Node *) _readJsonReturning();
	if (MATCH("JSONVALUEEXPR", 13))
		return (Node *) _readJsonValueExpr();
	if (MATCH("JSONCONSTRUCTOREXPR", 19))
		return (Node *) _readJsonConstructorExpr();
	if (MATCH("JSONISPREDICATE", 15))
		return (Node *) _readJsonIsPredicate();
	if (MATCH("JSONBEHAVIOR", 12))
		return (Node *) _readJsonBehavior();
	if (MATCH("JSONEXPR", 8))
		return (Node *) _readJsonExpr();
	if (MATCH("JSONTABLEPATH", 13))
		return (Node *) _readJsonTablePath();
	if (MATCH("JSONTABLEPATHSCAN", 17))
		return (Node *) _readJsonTablePathScan();
	if (MATCH("JSONTABLESIBLINGJOIN", 20))
		return (Node *) _readJsonTableSiblingJoin();
	if (MATCH("NULLTEST", 8))
		return (Node *) _readNullTest();
	if (MATCH("BOOLEANTEST", 11))
		return (Node *) _readBooleanTest();
	if (MATCH("MERGEACTION", 11))
		return (Node *) _readMergeAction();
	if (MATCH("COERCETODOMAIN", 14))
		return (Node *) _readCoerceToDomain();
	if (MATCH("COERCETODOMAINVALUE", 19))
		return (Node *) _readCoerceToDomainValue();
	if (MATCH("SETTODEFAULT", 12))
		return (Node *) _readSetToDefault();
	if (MATCH("CURRENTOFEXPR", 13))
		return (Node *) _readCurrentOfExpr();
	if (MATCH("NEXTVALUEEXPR", 13))
		return (Node *) _readNextValueExpr();
	if (MATCH("INFERENCEELEM", 13))
		return (Node *) _readInferenceElem();
	if (MATCH("RETURNINGEXPR", 13))
		return (Node *) _readReturningExpr();
	if (MATCH("TARGETENTRY", 11))
		return (Node *) _readTargetEntry();
	if (MATCH("RANGETBLREF", 11))
		return (Node *) _readRangeTblRef();
	if (MATCH("JOINEXPR", 8))
		return (Node *) _readJoinExpr();
	if (MATCH("FROMEXPR", 8))
		return (Node *) _readFromExpr();
	if (MATCH("ONCONFLICTEXPR", 14))
		return (Node *) _readOnConflictExpr();
	if (MATCH("QUERY", 5))
		return (Node *) _readQuery();
	if (MATCH("TYPENAME", 8))
		return (Node *) _readTypeName();
	if (MATCH("COLUMNREF", 9))
		return (Node *) _readColumnRef();
	if (MATCH("PARAMREF", 8))
		return (Node *) _readParamRef();
	if (MATCH("A_EXPR", 6))
		return (Node *) _readA_Expr();
	if (MATCH("A_CONST", 7))
		return (Node *) _readA_Const();
	if (MATCH("TYPECAST", 8))
		return (Node *) _readTypeCast();
	if (MATCH("COLLATECLAUSE", 13))
		return (Node *) _readCollateClause();
	if (MATCH("ROLESPEC", 8))
		return (Node *) _readRoleSpec();
	if (MATCH("FUNCCALL", 8))
		return (Node *) _readFuncCall();
	if (MATCH("A_STAR", 6))
		return (Node *) _readA_Star();
	if (MATCH("A_INDICES", 9))
		return (Node *) _readA_Indices();
	if (MATCH("A_INDIRECTION", 13))
		return (Node *) _readA_Indirection();
	if (MATCH("A_ARRAYEXPR", 11))
		return (Node *) _readA_ArrayExpr();
	if (MATCH("RESTARGET", 9))
		return (Node *) _readResTarget();
	if (MATCH("MULTIASSIGNREF", 14))
		return (Node *) _readMultiAssignRef();
	if (MATCH("SORTBY", 6))
		return (Node *) _readSortBy();
	if (MATCH("WINDOWDEF", 9))
		return (Node *) _readWindowDef();
	if (MATCH("RANGESUBSELECT", 14))
		return (Node *) _readRangeSubselect();
	if (MATCH("RANGEFUNCTION", 13))
		return (Node *) _readRangeFunction();
	if (MATCH("RANGETABLEFUNC", 14))
		return (Node *) _readRangeTableFunc();
	if (MATCH("RANGETABLEFUNCCOL", 17))
		return (Node *) _readRangeTableFuncCol();
	if (MATCH("RANGETABLESAMPLE", 16))
		return (Node *) _readRangeTableSample();
	if (MATCH("COLUMNDEF", 9))
		return (Node *) _readColumnDef();
	if (MATCH("TABLELIKECLAUSE", 15))
		return (Node *) _readTableLikeClause();
	if (MATCH("INDEXELEM", 9))
		return (Node *) _readIndexElem();
	if (MATCH("DEFELEM", 7))
		return (Node *) _readDefElem();
	if (MATCH("LOCKINGCLAUSE", 13))
		return (Node *) _readLockingClause();
	if (MATCH("XMLSERIALIZE", 12))
		return (Node *) _readXmlSerialize();
	if (MATCH("PARTITIONELEM", 13))
		return (Node *) _readPartitionElem();
	if (MATCH("PARTITIONSPEC", 13))
		return (Node *) _readPartitionSpec();
	if (MATCH("PARTITIONBOUNDSPEC", 18))
		return (Node *) _readPartitionBoundSpec();
	if (MATCH("PARTITIONRANGEDATUM", 19))
		return (Node *) _readPartitionRangeDatum();
	if (MATCH("PARTITIONCMD", 12))
		return (Node *) _readPartitionCmd();
	if (MATCH("RANGETBLENTRY", 13))
		return (Node *) _readRangeTblEntry();
	if (MATCH("RTEPERMISSIONINFO", 17))
		return (Node *) _readRTEPermissionInfo();
	if (MATCH("RANGETBLFUNCTION", 16))
		return (Node *) _readRangeTblFunction();
	if (MATCH("TABLESAMPLECLAUSE", 17))
		return (Node *) _readTableSampleClause();
	if (MATCH("WITHCHECKOPTION", 15))
		return (Node *) _readWithCheckOption();
	if (MATCH("SORTGROUPCLAUSE", 15))
		return (Node *) _readSortGroupClause();
	if (MATCH("GROUPINGSET", 11))
		return (Node *) _readGroupingSet();
	if (MATCH("WINDOWCLAUSE", 12))
		return (Node *) _readWindowClause();
	if (MATCH("ROWMARKCLAUSE", 13))
		return (Node *) _readRowMarkClause();
	if (MATCH("WITHCLAUSE", 10))
		return (Node *) _readWithClause();
	if (MATCH("INFERCLAUSE", 11))
		return (Node *) _readInferClause();
	if (MATCH("ONCONFLICTCLAUSE", 16))
		return (Node *) _readOnConflictClause();
	if (MATCH("CTESEARCHCLAUSE", 15))
		return (Node *) _readCTESearchClause();
	if (MATCH("CTECYCLECLAUSE", 14))
		return (Node *) _readCTECycleClause();
	if (MATCH("COMMONTABLEEXPR", 15))
		return (Node *) _readCommonTableExpr();
	if (MATCH("MERGEWHENCLAUSE", 15))
		return (Node *) _readMergeWhenClause();
	if (MATCH("RETURNINGOPTION", 15))
		return (Node *) _readReturningOption();
	if (MATCH("RETURNINGCLAUSE", 15))
		return (Node *) _readReturningClause();
	if (MATCH("TRIGGERTRANSITION", 17))
		return (Node *) _readTriggerTransition();
	if (MATCH("JSONOUTPUT", 10))
		return (Node *) _readJsonOutput();
	if (MATCH("JSONARGUMENT", 12))
		return (Node *) _readJsonArgument();
	if (MATCH("JSONFUNCEXPR", 12))
		return (Node *) _readJsonFuncExpr();
	if (MATCH("JSONTABLEPATHSPEC", 17))
		return (Node *) _readJsonTablePathSpec();
	if (MATCH("JSONTABLE", 9))
		return (Node *) _readJsonTable();
	if (MATCH("JSONTABLECOLUMN", 15))
		return (Node *) _readJsonTableColumn();
	if (MATCH("JSONKEYVALUE", 12))
		return (Node *) _readJsonKeyValue();
	if (MATCH("JSONPARSEEXPR", 13))
		return (Node *) _readJsonParseExpr();
	if (MATCH("JSONSCALAREXPR", 14))
		return (Node *) _readJsonScalarExpr();
	if (MATCH("JSONSERIALIZEEXPR", 17))
		return (Node *) _readJsonSerializeExpr();
	if (MATCH("JSONOBJECTCONSTRUCTOR", 21))
		return (Node *) _readJsonObjectConstructor();
	if (MATCH("JSONARRAYCONSTRUCTOR", 20))
		return (Node *) _readJsonArrayConstructor();
	if (MATCH("JSONARRAYQUERYCONSTRUCTOR", 25))
		return (Node *) _readJsonArrayQueryConstructor();
	if (MATCH("JSONAGGCONSTRUCTOR", 18))
		return (Node *) _readJsonAggConstructor();
	if (MATCH("JSONOBJECTAGG", 13))
		return (Node *) _readJsonObjectAgg();
	if (MATCH("JSONARRAYAGG", 12))
		return (Node *) _readJsonArrayAgg();
	if (MATCH("RAWSTMT", 7))
		return (Node *) _readRawStmt();
	if (MATCH("INSERTSTMT", 10))
		return (Node *) _readInsertStmt();
	if (MATCH("DELETESTMT", 10))
		return (Node *) _readDeleteStmt();
	if (MATCH("UPDATESTMT", 10))
		return (Node *) _readUpdateStmt();
	if (MATCH("MERGESTMT", 9))
		return (Node *) _readMergeStmt();
	if (MATCH("SELECTSTMT", 10))
		return (Node *) _readSelectStmt();
	if (MATCH("SETOPERATIONSTMT", 16))
		return (Node *) _readSetOperationStmt();
	if (MATCH("RETURNSTMT", 10))
		return (Node *) _readReturnStmt();
	if (MATCH("PLASSIGNSTMT", 12))
		return (Node *) _readPLAssignStmt();
	if (MATCH("CREATESCHEMASTMT", 16))
		return (Node *) _readCreateSchemaStmt();
	if (MATCH("ALTERTABLESTMT", 14))
		return (Node *) _readAlterTableStmt();
	if (MATCH("ALTERTABLECMD", 13))
		return (Node *) _readAlterTableCmd();
	if (MATCH("ATALTERCONSTRAINT", 17))
		return (Node *) _readATAlterConstraint();
	if (MATCH("REPLICAIDENTITYSTMT", 19))
		return (Node *) _readReplicaIdentityStmt();
	if (MATCH("ALTERCOLLATIONSTMT", 18))
		return (Node *) _readAlterCollationStmt();
	if (MATCH("ALTERDOMAINSTMT", 15))
		return (Node *) _readAlterDomainStmt();
	if (MATCH("GRANTSTMT", 9))
		return (Node *) _readGrantStmt();
	if (MATCH("OBJECTWITHARGS", 14))
		return (Node *) _readObjectWithArgs();
	if (MATCH("ACCESSPRIV", 10))
		return (Node *) _readAccessPriv();
	if (MATCH("GRANTROLESTMT", 13))
		return (Node *) _readGrantRoleStmt();
	if (MATCH("ALTERDEFAULTPRIVILEGESSTMT", 26))
		return (Node *) _readAlterDefaultPrivilegesStmt();
	if (MATCH("COPYSTMT", 8))
		return (Node *) _readCopyStmt();
	if (MATCH("VARIABLESETSTMT", 15))
		return (Node *) _readVariableSetStmt();
	if (MATCH("VARIABLESHOWSTMT", 16))
		return (Node *) _readVariableShowStmt();
	if (MATCH("CREATESTMT", 10))
		return (Node *) _readCreateStmt();
	if (MATCH("CONSTRAINT", 10))
		return (Node *) _readConstraint();
	if (MATCH("CREATETABLESPACESTMT", 20))
		return (Node *) _readCreateTableSpaceStmt();
	if (MATCH("DROPTABLESPACESTMT", 18))
		return (Node *) _readDropTableSpaceStmt();
	if (MATCH("ALTERTABLESPACEOPTIONSSTMT", 26))
		return (Node *) _readAlterTableSpaceOptionsStmt();
	if (MATCH("ALTERTABLEMOVEALLSTMT", 21))
		return (Node *) _readAlterTableMoveAllStmt();
	if (MATCH("CREATEEXTENSIONSTMT", 19))
		return (Node *) _readCreateExtensionStmt();
	if (MATCH("ALTEREXTENSIONSTMT", 18))
		return (Node *) _readAlterExtensionStmt();
	if (MATCH("ALTEREXTENSIONCONTENTSSTMT", 26))
		return (Node *) _readAlterExtensionContentsStmt();
	if (MATCH("CREATEFDWSTMT", 13))
		return (Node *) _readCreateFdwStmt();
	if (MATCH("ALTERFDWSTMT", 12))
		return (Node *) _readAlterFdwStmt();
	if (MATCH("CREATEFOREIGNSERVERSTMT", 23))
		return (Node *) _readCreateForeignServerStmt();
	if (MATCH("ALTERFOREIGNSERVERSTMT", 22))
		return (Node *) _readAlterForeignServerStmt();
	if (MATCH("CREATEFOREIGNTABLESTMT", 22))
		return (Node *) _readCreateForeignTableStmt();
	if (MATCH("CREATEUSERMAPPINGSTMT", 21))
		return (Node *) _readCreateUserMappingStmt();
	if (MATCH("ALTERUSERMAPPINGSTMT", 20))
		return (Node *) _readAlterUserMappingStmt();
	if (MATCH("DROPUSERMAPPINGSTMT", 19))
		return (Node *) _readDropUserMappingStmt();
	if (MATCH("IMPORTFOREIGNSCHEMASTMT", 23))
		return (Node *) _readImportForeignSchemaStmt();
	if (MATCH("CREATEPOLICYSTMT", 16))
		return (Node *) _readCreatePolicyStmt();
	if (MATCH("ALTERPOLICYSTMT", 15))
		return (Node *) _readAlterPolicyStmt();
	if (MATCH("CREATEAMSTMT", 12))
		return (Node *) _readCreateAmStmt();
	if (MATCH("CREATETRIGSTMT", 14))
		return (Node *) _readCreateTrigStmt();
	if (MATCH("CREATEEVENTTRIGSTMT", 19))
		return (Node *) _readCreateEventTrigStmt();
	if (MATCH("ALTEREVENTTRIGSTMT", 18))
		return (Node *) _readAlterEventTrigStmt();
	if (MATCH("CREATEPLANGSTMT", 15))
		return (Node *) _readCreatePLangStmt();
	if (MATCH("CREATEROLESTMT", 14))
		return (Node *) _readCreateRoleStmt();
	if (MATCH("ALTERROLESTMT", 13))
		return (Node *) _readAlterRoleStmt();
	if (MATCH("ALTERROLESETSTMT", 16))
		return (Node *) _readAlterRoleSetStmt();
	if (MATCH("DROPROLESTMT", 12))
		return (Node *) _readDropRoleStmt();
	if (MATCH("CREATESEQSTMT", 13))
		return (Node *) _readCreateSeqStmt();
	if (MATCH("ALTERSEQSTMT", 12))
		return (Node *) _readAlterSeqStmt();
	if (MATCH("DEFINESTMT", 10))
		return (Node *) _readDefineStmt();
	if (MATCH("CREATEDOMAINSTMT", 16))
		return (Node *) _readCreateDomainStmt();
	if (MATCH("CREATEOPCLASSSTMT", 17))
		return (Node *) _readCreateOpClassStmt();
	if (MATCH("CREATEOPCLASSITEM", 17))
		return (Node *) _readCreateOpClassItem();
	if (MATCH("CREATEOPFAMILYSTMT", 18))
		return (Node *) _readCreateOpFamilyStmt();
	if (MATCH("ALTEROPFAMILYSTMT", 17))
		return (Node *) _readAlterOpFamilyStmt();
	if (MATCH("DROPSTMT", 8))
		return (Node *) _readDropStmt();
	if (MATCH("TRUNCATESTMT", 12))
		return (Node *) _readTruncateStmt();
	if (MATCH("COMMENTSTMT", 11))
		return (Node *) _readCommentStmt();
	if (MATCH("SECLABELSTMT", 12))
		return (Node *) _readSecLabelStmt();
	if (MATCH("DECLARECURSORSTMT", 17))
		return (Node *) _readDeclareCursorStmt();
	if (MATCH("CLOSEPORTALSTMT", 15))
		return (Node *) _readClosePortalStmt();
	if (MATCH("FETCHSTMT", 9))
		return (Node *) _readFetchStmt();
	if (MATCH("INDEXSTMT", 9))
		return (Node *) _readIndexStmt();
	if (MATCH("CREATESTATSSTMT", 15))
		return (Node *) _readCreateStatsStmt();
	if (MATCH("STATSELEM", 9))
		return (Node *) _readStatsElem();
	if (MATCH("ALTERSTATSSTMT", 14))
		return (Node *) _readAlterStatsStmt();
	if (MATCH("CREATEFUNCTIONSTMT", 18))
		return (Node *) _readCreateFunctionStmt();
	if (MATCH("FUNCTIONPARAMETER", 17))
		return (Node *) _readFunctionParameter();
	if (MATCH("ALTERFUNCTIONSTMT", 17))
		return (Node *) _readAlterFunctionStmt();
	if (MATCH("DOSTMT", 6))
		return (Node *) _readDoStmt();
	if (MATCH("CALLSTMT", 8))
		return (Node *) _readCallStmt();
	if (MATCH("RENAMESTMT", 10))
		return (Node *) _readRenameStmt();
	if (MATCH("ALTEROBJECTDEPENDSSTMT", 22))
		return (Node *) _readAlterObjectDependsStmt();
	if (MATCH("ALTEROBJECTSCHEMASTMT", 21))
		return (Node *) _readAlterObjectSchemaStmt();
	if (MATCH("ALTEROWNERSTMT", 14))
		return (Node *) _readAlterOwnerStmt();
	if (MATCH("ALTEROPERATORSTMT", 17))
		return (Node *) _readAlterOperatorStmt();
	if (MATCH("ALTERTYPESTMT", 13))
		return (Node *) _readAlterTypeStmt();
	if (MATCH("RULESTMT", 8))
		return (Node *) _readRuleStmt();
	if (MATCH("NOTIFYSTMT", 10))
		return (Node *) _readNotifyStmt();
	if (MATCH("LISTENSTMT", 10))
		return (Node *) _readListenStmt();
	if (MATCH("UNLISTENSTMT", 12))
		return (Node *) _readUnlistenStmt();
	if (MATCH("TRANSACTIONSTMT", 15))
		return (Node *) _readTransactionStmt();
	if (MATCH("COMPOSITETYPESTMT", 17))
		return (Node *) _readCompositeTypeStmt();
	if (MATCH("CREATEENUMSTMT", 14))
		return (Node *) _readCreateEnumStmt();
	if (MATCH("CREATERANGESTMT", 15))
		return (Node *) _readCreateRangeStmt();
	if (MATCH("ALTERENUMSTMT", 13))
		return (Node *) _readAlterEnumStmt();
	if (MATCH("VIEWSTMT", 8))
		return (Node *) _readViewStmt();
	if (MATCH("LOADSTMT", 8))
		return (Node *) _readLoadStmt();
	if (MATCH("CREATEDBSTMT", 12))
		return (Node *) _readCreatedbStmt();
	if (MATCH("ALTERDATABASESTMT", 17))
		return (Node *) _readAlterDatabaseStmt();
	if (MATCH("ALTERDATABASEREFRESHCOLLSTMT", 28))
		return (Node *) _readAlterDatabaseRefreshCollStmt();
	if (MATCH("ALTERDATABASESETSTMT", 20))
		return (Node *) _readAlterDatabaseSetStmt();
	if (MATCH("DROPDBSTMT", 10))
		return (Node *) _readDropdbStmt();
	if (MATCH("ALTERSYSTEMSTMT", 15))
		return (Node *) _readAlterSystemStmt();
	if (MATCH("CLUSTERSTMT", 11))
		return (Node *) _readClusterStmt();
	if (MATCH("VACUUMSTMT", 10))
		return (Node *) _readVacuumStmt();
	if (MATCH("VACUUMRELATION", 14))
		return (Node *) _readVacuumRelation();
	if (MATCH("EXPLAINSTMT", 11))
		return (Node *) _readExplainStmt();
	if (MATCH("CREATETABLEASSTMT", 17))
		return (Node *) _readCreateTableAsStmt();
	if (MATCH("REFRESHMATVIEWSTMT", 18))
		return (Node *) _readRefreshMatViewStmt();
	if (MATCH("CHECKPOINTSTMT", 14))
		return (Node *) _readCheckPointStmt();
	if (MATCH("DISCARDSTMT", 11))
		return (Node *) _readDiscardStmt();
	if (MATCH("LOCKSTMT", 8))
		return (Node *) _readLockStmt();
	if (MATCH("CONSTRAINTSSETSTMT", 18))
		return (Node *) _readConstraintsSetStmt();
	if (MATCH("REINDEXSTMT", 11))
		return (Node *) _readReindexStmt();
	if (MATCH("CREATECONVERSIONSTMT", 20))
		return (Node *) _readCreateConversionStmt();
	if (MATCH("CREATECASTSTMT", 14))
		return (Node *) _readCreateCastStmt();
	if (MATCH("CREATETRANSFORMSTMT", 19))
		return (Node *) _readCreateTransformStmt();
	if (MATCH("PREPARESTMT", 11))
		return (Node *) _readPrepareStmt();
	if (MATCH("EXECUTESTMT", 11))
		return (Node *) _readExecuteStmt();
	if (MATCH("DEALLOCATESTMT", 14))
		return (Node *) _readDeallocateStmt();
	if (MATCH("DROPOWNEDSTMT", 13))
		return (Node *) _readDropOwnedStmt();
	if (MATCH("REASSIGNOWNEDSTMT", 17))
		return (Node *) _readReassignOwnedStmt();
	if (MATCH("ALTERTSDICTIONARYSTMT", 21))
		return (Node *) _readAlterTSDictionaryStmt();
	if (MATCH("ALTERTSCONFIGURATIONSTMT", 24))
		return (Node *) _readAlterTSConfigurationStmt();
	if (MATCH("PUBLICATIONTABLE", 16))
		return (Node *) _readPublicationTable();
	if (MATCH("PUBLICATIONOBJSPEC", 18))
		return (Node *) _readPublicationObjSpec();
	if (MATCH("CREATEPUBLICATIONSTMT", 21))
		return (Node *) _readCreatePublicationStmt();
	if (MATCH("ALTERPUBLICATIONSTMT", 20))
		return (Node *) _readAlterPublicationStmt();
	if (MATCH("CREATESUBSCRIPTIONSTMT", 22))
		return (Node *) _readCreateSubscriptionStmt();
	if (MATCH("ALTERSUBSCRIPTIONSTMT", 21))
		return (Node *) _readAlterSubscriptionStmt();
	if (MATCH("DROPSUBSCRIPTIONSTMT", 20))
		return (Node *) _readDropSubscriptionStmt();
	if (MATCH("GROUPBYORDERING", 15))
		return (Node *) _readGroupByOrdering();
	if (MATCH("PLACEHOLDERVAR", 14))
		return (Node *) _readPlaceHolderVar();
	if (MATCH("APPENDRELINFO", 13))
		return (Node *) _readAppendRelInfo();
	if (MATCH("PLANNEDSTMT", 11))
		return (Node *) _readPlannedStmt();
	if (MATCH("RESULT", 6))
		return (Node *) _readResult();
	if (MATCH("PROJECTSET", 10))
		return (Node *) _readProjectSet();
	if (MATCH("MODIFYTABLE", 11))
		return (Node *) _readModifyTable();
	if (MATCH("APPEND", 6))
		return (Node *) _readAppend();
	if (MATCH("MERGEAPPEND", 11))
		return (Node *) _readMergeAppend();
	if (MATCH("RECURSIVEUNION", 14))
		return (Node *) _readRecursiveUnion();
	if (MATCH("BITMAPAND", 9))
		return (Node *) _readBitmapAnd();
	if (MATCH("BITMAPOR", 8))
		return (Node *) _readBitmapOr();
	if (MATCH("SEQSCAN", 7))
		return (Node *) _readSeqScan();
	if (MATCH("SAMPLESCAN", 10))
		return (Node *) _readSampleScan();
	if (MATCH("INDEXSCAN", 9))
		return (Node *) _readIndexScan();
	if (MATCH("INDEXONLYSCAN", 13))
		return (Node *) _readIndexOnlyScan();
	if (MATCH("BITMAPINDEXSCAN", 15))
		return (Node *) _readBitmapIndexScan();
	if (MATCH("BITMAPHEAPSCAN", 14))
		return (Node *) _readBitmapHeapScan();
	if (MATCH("TIDSCAN", 7))
		return (Node *) _readTidScan();
	if (MATCH("TIDRANGESCAN", 12))
		return (Node *) _readTidRangeScan();
	if (MATCH("SUBQUERYSCAN", 12))
		return (Node *) _readSubqueryScan();
	if (MATCH("FUNCTIONSCAN", 12))
		return (Node *) _readFunctionScan();
	if (MATCH("VALUESSCAN", 10))
		return (Node *) _readValuesScan();
	if (MATCH("TABLEFUNCSCAN", 13))
		return (Node *) _readTableFuncScan();
	if (MATCH("CTESCAN", 7))
		return (Node *) _readCteScan();
	if (MATCH("NAMEDTUPLESTORESCAN", 19))
		return (Node *) _readNamedTuplestoreScan();
	if (MATCH("WORKTABLESCAN", 13))
		return (Node *) _readWorkTableScan();
	if (MATCH("FOREIGNSCAN", 11))
		return (Node *) _readForeignScan();
	if (MATCH("CUSTOMSCAN", 10))
		return (Node *) _readCustomScan();
	if (MATCH("NESTLOOP", 8))
		return (Node *) _readNestLoop();
	if (MATCH("NESTLOOPPARAM", 13))
		return (Node *) _readNestLoopParam();
	if (MATCH("MERGEJOIN", 9))
		return (Node *) _readMergeJoin();
	if (MATCH("HASHJOIN", 8))
		return (Node *) _readHashJoin();
	if (MATCH("MATERIAL", 8))
		return (Node *) _readMaterial();
	if (MATCH("MEMOIZE", 7))
		return (Node *) _readMemoize();
	if (MATCH("SORT", 4))
		return (Node *) _readSort();
	if (MATCH("INCREMENTALSORT", 15))
		return (Node *) _readIncrementalSort();
	if (MATCH("GROUP", 5))
		return (Node *) _readGroup();
	if (MATCH("AGG", 3))
		return (Node *) _readAgg();
	if (MATCH("WINDOWAGG", 9))
		return (Node *) _readWindowAgg();
	if (MATCH("UNIQUE", 6))
		return (Node *) _readUnique();
	if (MATCH("GATHER", 6))
		return (Node *) _readGather();
	if (MATCH("GATHERMERGE", 11))
		return (Node *) _readGatherMerge();
	if (MATCH("HASH", 4))
		return (Node *) _readHash();
	if (MATCH("SETOP", 5))
		return (Node *) _readSetOp();
	if (MATCH("LOCKROWS", 8))
		return (Node *) _readLockRows();
	if (MATCH("LIMIT", 5))
		return (Node *) _readLimit();
	if (MATCH("PLANROWMARK", 11))
		return (Node *) _readPlanRowMark();
	if (MATCH("PARTITIONPRUNEINFO", 18))
		return (Node *) _readPartitionPruneInfo();
	if (MATCH("PARTITIONEDRELPRUNEINFO", 23))
		return (Node *) _readPartitionedRelPruneInfo();
	if (MATCH("PARTITIONPRUNESTEPOP", 20))
		return (Node *) _readPartitionPruneStepOp();
	if (MATCH("PARTITIONPRUNESTEPCOMBINE", 25))
		return (Node *) _readPartitionPruneStepCombine();
	if (MATCH("PLANINVALITEM", 13))
		return (Node *) _readPlanInvalItem();
	if (MATCH("EXTENSIBLENODE", 14))
		return (Node *) _readExtensibleNode();
