/*
 * This file was generated automatically by ExtUtils::ParseXS version 3.57 from the
 * contents of Util.xs. Do not edit this file, edit Util.xs instead.
 *
 *    ANY CHANGES MADE HERE WILL BE LOST!
 *
 */

#line 1 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
/**********************************************************************
 * PostgreSQL::InServer::Util
 *
 * src/pl/plperl/Util.xs
 *
 * Defines plperl interfaces for general-purpose utilities.
 * This module is bootstrapped as soon as an interpreter is initialized.
 * Currently doesn't define a PACKAGE= so all subs are in main:: to avoid
 * the need for explicit importing.
 *
 **********************************************************************/

/* this must be first: */
#include "postgres.h"

#include "fmgr.h"
#include "utils/builtins.h"
#include "utils/bytea.h"       /* for byteain & byteaout */
#include "varatt.h"

/* perl stuff */
#define PG_NEED_PERL_XSUB_H
#include "plperl.h"


static text *
sv2text(SV *sv)
{
	char	   *str = sv2cstr(sv);
	text	   *text;

	text = cstring_to_text(str);
	pfree(str);
	return text;
}

#line 47 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
#ifndef PERL_UNUSED_VAR
#  define PERL_UNUSED_VAR(var) if (0) var = var
#endif

#ifndef dVAR
#  define dVAR		dNOOP
#endif


/* This stuff is not part of the API! You have been warned. */
#ifndef PERL_VERSION_DECIMAL
#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
#endif
#ifndef PERL_DECIMAL_VERSION
#  define PERL_DECIMAL_VERSION \
	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
#endif
#ifndef PERL_VERSION_GE
#  define PERL_VERSION_GE(r,v,s) \
	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
#endif
#ifndef PERL_VERSION_LE
#  define PERL_VERSION_LE(r,v,s) \
	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
#endif

/* XS_INTERNAL is the explicit static-linkage variant of the default
 * XS macro.
 *
 * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
 * "STATIC", ie. it exports XSUB symbols. You probably don't want that
 * for anything but the BOOT XSUB.
 *
 * See XSUB.h in core!
 */


/* TODO: This might be compatible further back than 5.10.0. */
#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
#  undef XS_EXTERNAL
#  undef XS_INTERNAL
#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
#    define XS_INTERNAL(name) STATIC XSPROTO(name)
#  endif
#  if defined(__SYMBIAN32__)
#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
#  endif
#  ifndef XS_EXTERNAL
#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
#    else
#      ifdef __cplusplus
#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
#        define XS_INTERNAL(name) static XSPROTO(name)
#      else
#        define XS_EXTERNAL(name) XSPROTO(name)
#        define XS_INTERNAL(name) STATIC XSPROTO(name)
#      endif
#    endif
#  endif
#endif

/* perl >= 5.10.0 && perl <= 5.15.1 */


/* The XS_EXTERNAL macro is used for functions that must not be static
 * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
 * macro defined, the best we can do is assume XS is the same.
 * Dito for XS_INTERNAL.
 */
#ifndef XS_EXTERNAL
#  define XS_EXTERNAL(name) XS(name)
#endif
#ifndef XS_INTERNAL
#  define XS_INTERNAL(name) XS(name)
#endif

/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
 * internal macro that we're free to redefine for varying linkage due
 * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
 * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
 */

#undef XS_EUPXS
#if defined(PERL_EUPXS_ALWAYS_EXPORT)
#  define XS_EUPXS(name) XS_EXTERNAL(name)
#else
   /* default to internal */
#  define XS_EUPXS(name) XS_INTERNAL(name)
#endif

#ifndef PERL_ARGS_ASSERT_CROAK_XS_USAGE
#define PERL_ARGS_ASSERT_CROAK_XS_USAGE assert(cv); assert(params)

/* prototype to pass -Wmissing-prototypes */
STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params);

STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params)
{
    const GV *const gv = CvGV(cv);

    PERL_ARGS_ASSERT_CROAK_XS_USAGE;

    if (gv) {
        const char *const gvname = GvNAME(gv);
        const HV *const stash = GvSTASH(gv);
        const char *const hvname = stash ? HvNAME(stash) : NULL;

        if (hvname)
	    Perl_croak_nocontext("Usage: %s::%s(%s)", hvname, gvname, params);
        else
	    Perl_croak_nocontext("Usage: %s(%s)", gvname, params);
    } else {
        /* Pants. I don't think that it should be possible to get here. */
	Perl_croak_nocontext("Usage: CODE(0x%" UVxf ")(%s)", PTR2UV(cv), params);
    }
}
#undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE

#define croak_xs_usage        S_croak_xs_usage

#endif

/* NOTE: the prototype of newXSproto() is different in versions of perls,
 * so we define a portable version of newXSproto()
 */
#ifdef newXS_flags
#define newXSproto_portable(name, c_impl, file, proto) newXS_flags(name, c_impl, file, proto, 0)
#else
#define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
#endif /* !defined(newXS_flags) */

#if PERL_VERSION_LE(5, 21, 5)
#  define newXS_deffile(a,b) Perl_newXS(aTHX_ a,b,file)
#else
#  define newXS_deffile(a,b) Perl_newXS_deffile(aTHX_ a,b)
#endif

/* simple backcompat versions of the TARGx() macros with no optimisation */
#ifndef TARGi
#  define TARGi(iv, do_taint) sv_setiv_mg(TARG, iv)
#  define TARGu(uv, do_taint) sv_setuv_mg(TARG, uv)
#  define TARGn(nv, do_taint) sv_setnv_mg(TARG, nv)
#endif

#line 198 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"

XS_EUPXS(XS___aliased_constants); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS___aliased_constants)
{
    dVAR; dXSARGS;
    dXSI32;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
	int	RETVAL;
	dXSTARG;
#line 53 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
    /* uses the ALIAS value as the return value */
    RETVAL = ix;
#line 213 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
	TARGi((IV)RETVAL, 1);
	ST(0) = TARG;
    }
    XSRETURN(1);
}


XS_EUPXS(XS__elog); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__elog)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "level, msg");
    {
	int	level = (int)SvIV(ST(0))
;
	SV *	msg = ST(1)
;
#line 64 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
        if (level > ERROR)      /* no PANIC allowed thanks */
            level = ERROR;
        if (level < DEBUG5)
            level = DEBUG5;
        plperl_util_elog(level, msg);
#line 238 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS__quote_literal); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__quote_literal)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "sv");
    {
	SV *	sv = ST(0)
;
	SV *	RETVAL;
#line 74 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
    if (!sv || !SvOK(sv)) {
        RETVAL = &PL_sv_undef;
    }
    else {
        text *arg = sv2text(sv);
		text *quoted = DatumGetTextPP(DirectFunctionCall1(quote_literal, PointerGetDatum(arg)));
		char *str;

		pfree(arg);
		str = text_to_cstring(quoted);
		RETVAL = cstr2sv(str);
		pfree(str);
    }
#line 268 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS__quote_nullable); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__quote_nullable)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "sv");
    {
	SV *	sv = ST(0)
;
	SV *	RETVAL;
#line 94 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
    if (!sv || !SvOK(sv))
	{
        RETVAL = cstr2sv("NULL");
    }
    else
	{
        text *arg = sv2text(sv);
		text *quoted = DatumGetTextPP(DirectFunctionCall1(quote_nullable, PointerGetDatum(arg)));
		char *str;

		pfree(arg);
		str = text_to_cstring(quoted);
		RETVAL = cstr2sv(str);
		pfree(str);
    }
#line 302 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS__quote_ident); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__quote_ident)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "sv");
    {
	SV *	sv = ST(0)
;
#line 116 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
        text *arg;
		text *quoted;
		char *str;
#line 323 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
	SV *	RETVAL;
#line 120 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
        arg = sv2text(sv);
		quoted = DatumGetTextPP(DirectFunctionCall1(quote_ident, PointerGetDatum(arg)));

		pfree(arg);
		str = text_to_cstring(quoted);
		RETVAL = cstr2sv(str);
		pfree(str);
#line 333 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS__decode_bytea); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__decode_bytea)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "sv");
    {
	SV *	sv = ST(0)
;
#line 134 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
        char *arg;
        text *ret;
#line 353 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
	SV *	RETVAL;
#line 137 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
        arg = SvPVbyte_nolen(sv);
        ret = DatumGetTextPP(DirectFunctionCall1(byteain, PointerGetDatum(arg)));
        /* not cstr2sv because this is raw bytes not utf8'able */
        RETVAL = newSVpvn(VARDATA_ANY(ret), VARSIZE_ANY_EXHDR(ret));
#line 360 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS__encode_bytea); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__encode_bytea)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "sv");
    {
	SV *	sv = ST(0)
;
#line 148 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
        text *arg;
        char *ret;
		STRLEN len;
#line 381 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
	SV *	RETVAL;
#line 152 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
        /* not sv2text because this is raw bytes not utf8'able */
        ret = SvPVbyte(sv, len);
		arg = cstring_to_text_with_len(ret, len);
        ret = DatumGetCString(DirectFunctionCall1(byteaout, PointerGetDatum(arg)));
        RETVAL = cstr2sv(ret);
#line 389 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS__looks_like_number); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__looks_like_number)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "sv");
    {
	SV *	sv = ST(0)
;
	SV *	RETVAL;
#line 164 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
    if (!SvOK(sv))
        RETVAL = &PL_sv_undef;
    else if ( looks_like_number(sv) )
        RETVAL = &PL_sv_yes;
    else
        RETVAL = &PL_sv_no;
#line 414 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS__encode_typed_literal); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__encode_typed_literal)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "sv, typname");
    {
	SV *	sv = ST(0)
;
	char *	typname = (char *)SvPV_nolen(ST(1))
;
#line 178 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
		char 	*outstr;
#line 435 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
	SV *	RETVAL;
#line 180 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
		outstr = plperl_sv_to_literal(sv, typname);
		if (outstr == NULL)
			RETVAL = &PL_sv_undef;
		else
			RETVAL = cstr2sv(outstr);
#line 443 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}

#ifdef __cplusplus
extern "C" {
#endif
XS_EXTERNAL(boot_PostgreSQL__InServer__Util); /* prototype to pass -Wmissing-prototypes */
XS_EXTERNAL(boot_PostgreSQL__InServer__Util)
{
#if PERL_VERSION_LE(5, 21, 5)
    dVAR; dXSARGS;
#else
    dVAR; dXSBOOTARGSAPIVERCHK;
#endif
#if PERL_VERSION_LE(5, 8, 999) /* PERL_VERSION_LT is 5.33+ */
    char* file = __FILE__;
#else
    const char* file = __FILE__;
#endif

    PERL_UNUSED_VAR(file);

    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
#if PERL_VERSION_LE(5, 21, 5) && defined(XS_APIVERSION_BOOTCHECK)
  XS_APIVERSION_BOOTCHECK;
#endif

        cv = newXSproto_portable("DEBUG", XS___aliased_constants, file, "");
        XSANY.any_i32 = DEBUG2;
        cv = newXSproto_portable("ERROR", XS___aliased_constants, file, "");
        XSANY.any_i32 = ERROR;
        cv = newXSproto_portable("INFO", XS___aliased_constants, file, "");
        XSANY.any_i32 = INFO;
        cv = newXSproto_portable("LOG", XS___aliased_constants, file, "");
        XSANY.any_i32 = LOG;
        cv = newXSproto_portable("NOTICE", XS___aliased_constants, file, "");
        XSANY.any_i32 = NOTICE;
        cv = newXSproto_portable("WARNING", XS___aliased_constants, file, "");
        XSANY.any_i32 = WARNING;
        cv = newXSproto_portable("_aliased_constants", XS___aliased_constants, file, "");
        XSANY.any_i32 = 0;
        (void)newXSproto_portable("elog", XS__elog, file, "$$");
        (void)newXSproto_portable("quote_literal", XS__quote_literal, file, "$");
        (void)newXSproto_portable("quote_nullable", XS__quote_nullable, file, "$");
        (void)newXSproto_portable("quote_ident", XS__quote_ident, file, "$");
        (void)newXSproto_portable("decode_bytea", XS__decode_bytea, file, "$");
        (void)newXSproto_portable("encode_bytea", XS__encode_bytea, file, "$");
        (void)newXSproto_portable("looks_like_number", XS__looks_like_number, file, "$");
        (void)newXSproto_portable("encode_typed_literal", XS__encode_typed_literal, file, "$$");

    /* Initialisation Section */

#line 189 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.xs"
    items = 0;  /* avoid 'unused variable' warning */

#line 503 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/Util.c"

    /* End of Initialisation Section */

#if PERL_VERSION_LE(5, 21, 5)
#  if PERL_VERSION_GE(5, 9, 0)
    if (PL_unitcheckav)
        call_list(PL_scopestack_ix, PL_unitcheckav);
#  endif
    XSRETURN_YES;
#else
    Perl_xs_boot_epilog(aTHX_ ax);
#endif
}

#ifdef __cplusplus
}
#endif
