/*
 * This file was generated automatically by ExtUtils::ParseXS version 3.57 from the
 * contents of SPI.xs. Do not edit this file, edit SPI.xs instead.
 *
 *    ANY CHANGES MADE HERE WILL BE LOST!
 *
 */

#line 1 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
/**********************************************************************
 * PostgreSQL::InServer::SPI
 *
 * SPI interface for plperl.
 *
 *    src/pl/plperl/SPI.xs
 *
 **********************************************************************/

/* this must be first: */
#include "postgres.h"

/* perl stuff */
#define PG_NEED_PERL_XSUB_H
#include "plperl.h"


#line 28 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
#ifndef PERL_UNUSED_VAR
#  define PERL_UNUSED_VAR(var) if (0) var = var
#endif

#ifndef dVAR
#  define dVAR		dNOOP
#endif


/* This stuff is not part of the API! You have been warned. */
#ifndef PERL_VERSION_DECIMAL
#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
#endif
#ifndef PERL_DECIMAL_VERSION
#  define PERL_DECIMAL_VERSION \
	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
#endif
#ifndef PERL_VERSION_GE
#  define PERL_VERSION_GE(r,v,s) \
	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
#endif
#ifndef PERL_VERSION_LE
#  define PERL_VERSION_LE(r,v,s) \
	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
#endif

/* XS_INTERNAL is the explicit static-linkage variant of the default
 * XS macro.
 *
 * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
 * "STATIC", ie. it exports XSUB symbols. You probably don't want that
 * for anything but the BOOT XSUB.
 *
 * See XSUB.h in core!
 */


/* TODO: This might be compatible further back than 5.10.0. */
#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
#  undef XS_EXTERNAL
#  undef XS_INTERNAL
#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
#    define XS_INTERNAL(name) STATIC XSPROTO(name)
#  endif
#  if defined(__SYMBIAN32__)
#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
#  endif
#  ifndef XS_EXTERNAL
#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
#    else
#      ifdef __cplusplus
#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
#        define XS_INTERNAL(name) static XSPROTO(name)
#      else
#        define XS_EXTERNAL(name) XSPROTO(name)
#        define XS_INTERNAL(name) STATIC XSPROTO(name)
#      endif
#    endif
#  endif
#endif

/* perl >= 5.10.0 && perl <= 5.15.1 */


/* The XS_EXTERNAL macro is used for functions that must not be static
 * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
 * macro defined, the best we can do is assume XS is the same.
 * Dito for XS_INTERNAL.
 */
#ifndef XS_EXTERNAL
#  define XS_EXTERNAL(name) XS(name)
#endif
#ifndef XS_INTERNAL
#  define XS_INTERNAL(name) XS(name)
#endif

/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
 * internal macro that we're free to redefine for varying linkage due
 * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
 * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
 */

#undef XS_EUPXS
#if defined(PERL_EUPXS_ALWAYS_EXPORT)
#  define XS_EUPXS(name) XS_EXTERNAL(name)
#else
   /* default to internal */
#  define XS_EUPXS(name) XS_INTERNAL(name)
#endif

#ifndef PERL_ARGS_ASSERT_CROAK_XS_USAGE
#define PERL_ARGS_ASSERT_CROAK_XS_USAGE assert(cv); assert(params)

/* prototype to pass -Wmissing-prototypes */
STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params);

STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params)
{
    const GV *const gv = CvGV(cv);

    PERL_ARGS_ASSERT_CROAK_XS_USAGE;

    if (gv) {
        const char *const gvname = GvNAME(gv);
        const HV *const stash = GvSTASH(gv);
        const char *const hvname = stash ? HvNAME(stash) : NULL;

        if (hvname)
	    Perl_croak_nocontext("Usage: %s::%s(%s)", hvname, gvname, params);
        else
	    Perl_croak_nocontext("Usage: %s(%s)", gvname, params);
    } else {
        /* Pants. I don't think that it should be possible to get here. */
	Perl_croak_nocontext("Usage: CODE(0x%" UVxf ")(%s)", PTR2UV(cv), params);
    }
}
#undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE

#define croak_xs_usage        S_croak_xs_usage

#endif

/* NOTE: the prototype of newXSproto() is different in versions of perls,
 * so we define a portable version of newXSproto()
 */
#ifdef newXS_flags
#define newXSproto_portable(name, c_impl, file, proto) newXS_flags(name, c_impl, file, proto, 0)
#else
#define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
#endif /* !defined(newXS_flags) */

#if PERL_VERSION_LE(5, 21, 5)
#  define newXS_deffile(a,b) Perl_newXS(aTHX_ a,b,file)
#else
#  define newXS_deffile(a,b) Perl_newXS_deffile(aTHX_ a,b)
#endif

/* simple backcompat versions of the TARGx() macros with no optimisation */
#ifndef TARGi
#  define TARGi(iv, do_taint) sv_setiv_mg(TARG, iv)
#  define TARGu(uv, do_taint) sv_setuv_mg(TARG, uv)
#  define TARGn(nv, do_taint) sv_setnv_mg(TARG, nv)
#endif

#line 179 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"

XS_EUPXS(XS__spi_exec_query); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__spi_exec_query)
{
    dVAR; dXSARGS;
    if (items < 1)
       croak_xs_usage(cv,  "sv, ...");
    {
	SV*	sv = ST(0)
;
#line 27 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
		HV *ret_hash;
		int limit = 0;
		char *query;
#line 194 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
	SV *	RETVAL;
#line 31 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
		if (items > 2)
			croak("Usage: spi_exec_query(query, limit) "
				  "or spi_exec_query(query)");
		if (items == 2)
			limit = SvIV(ST(1));
		query = sv2cstr(sv);
		ret_hash = plperl_spi_exec(query, limit);
		pfree(query);
		RETVAL = newRV_noinc((SV*) ret_hash);
#line 206 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS__return_next); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__return_next)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "rv");
    {
	SV *	rv = ST(0)
;
#line 47 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
		plperl_return_next(rv);
#line 225 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS__spi_query); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__spi_query)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "sv");
    {
	SV *	sv = ST(0)
;
	SV *	RETVAL;
#line 53 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
		char* query = sv2cstr(sv);
		RETVAL = plperl_spi_query(query);
		pfree(query);
#line 245 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS__spi_fetchrow); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__spi_fetchrow)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "sv");
    {
	SV*	sv = ST(0)
;
	SV *	RETVAL;
#line 63 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
		char* cursor = sv2cstr(sv);
		RETVAL = plperl_spi_fetchrow(cursor);
		pfree(cursor);
#line 267 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS__spi_prepare); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__spi_prepare)
{
    dVAR; dXSARGS;
    if (items < 1)
       croak_xs_usage(cv,  "sv, ...");
    {
	SV*	sv = ST(0)
;
	SV *	RETVAL;
#line 73 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
		int i;
		SV** argv;
		char* query = sv2cstr(sv);
		if (items < 1)
			Perl_croak(aTHX_ "Usage: spi_prepare(query, ...)");
		argv = palloc_array(SV*, items - 1);
		for ( i = 1; i < items; i++)
			argv[i - 1] = ST(i);
		RETVAL = plperl_spi_prepare(query, items - 1, argv);
		pfree( argv);
		pfree(query);
#line 297 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS__spi_exec_prepared); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__spi_exec_prepared)
{
    dVAR; dXSARGS;
    if (items < 1)
       croak_xs_usage(cv,  "sv, ...");
    {
	SV*	sv = ST(0)
;
#line 91 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
		HV *ret_hash;
#line 316 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
	SV *	RETVAL;
#line 93 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
		HV *attr = NULL;
		int i, offset = 1, argc;
		SV ** argv;
		char *query = sv2cstr(sv);
		if ( items < 1)
			Perl_croak(aTHX_ "Usage: spi_exec_prepared(query, [\\%%attr,] "
					   "[\\@bind_values])");
		if ( items > 1 && SvROK( ST( 1)) && SvTYPE( SvRV( ST( 1))) == SVt_PVHV)
		{
			attr = ( HV*) SvRV(ST(1));
			offset++;
		}
		argc = items - offset;
		argv = palloc_array(SV*, argc);
		for ( i = 0; offset < items; offset++, i++)
			argv[i] = ST(offset);
		ret_hash = plperl_spi_exec_prepared(query, attr, argc, argv);
		RETVAL = newRV_noinc((SV*)ret_hash);
		pfree( argv);
		pfree(query);
#line 339 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS__spi_query_prepared); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__spi_query_prepared)
{
    dVAR; dXSARGS;
    if (items < 1)
       croak_xs_usage(cv,  "sv, ...");
    {
	SV *	sv = ST(0)
;
	SV *	RETVAL;
#line 120 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
		int i;
		SV ** argv;
		char *query = sv2cstr(sv);
		if ( items < 1)
			Perl_croak(aTHX_ "Usage: spi_query_prepared(query, "
					   "[\\@bind_values])");
		argv = palloc_array(SV*, items - 1);
		for ( i = 1; i < items; i++)
			argv[i - 1] = ST(i);
		RETVAL = plperl_spi_query_prepared(query, items - 1, argv);
		pfree( argv);
		pfree(query);
#line 370 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS__spi_freeplan); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__spi_freeplan)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "sv");
    {
	SV *	sv = ST(0)
;
#line 139 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
		char *query = sv2cstr(sv);
		plperl_spi_freeplan(query);
		pfree(query);
#line 391 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS__spi_cursor_close); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__spi_cursor_close)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "sv");
    {
	SV *	sv = ST(0)
;
#line 147 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
		char *cursor = sv2cstr(sv);
		plperl_spi_cursor_close(cursor);
		pfree(cursor);
#line 410 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS__spi_commit); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__spi_commit)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
#line 154 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
		plperl_spi_commit();
#line 425 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS__spi_rollback); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS__spi_rollback)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
#line 159 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
		plperl_spi_rollback();
#line 440 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"
    }
    XSRETURN_EMPTY;
}

#ifdef __cplusplus
extern "C" {
#endif
XS_EXTERNAL(boot_PostgreSQL__InServer__SPI); /* prototype to pass -Wmissing-prototypes */
XS_EXTERNAL(boot_PostgreSQL__InServer__SPI)
{
#if PERL_VERSION_LE(5, 21, 5)
    dVAR; dXSARGS;
#else
    dVAR; dXSBOOTARGSAPIVERCHK;
#endif
#if PERL_VERSION_LE(5, 8, 999) /* PERL_VERSION_LT is 5.33+ */
    char* file = __FILE__;
#else
    const char* file = __FILE__;
#endif

    PERL_UNUSED_VAR(file);

    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
#if PERL_VERSION_LE(5, 21, 5) && defined(XS_APIVERSION_BOOTCHECK)
  XS_APIVERSION_BOOTCHECK;
#endif

        (void)newXSproto_portable("spi_exec_query", XS__spi_exec_query, file, "$;@");
        (void)newXSproto_portable("return_next", XS__return_next, file, "$");
        (void)newXSproto_portable("spi_query", XS__spi_query, file, "$");
        (void)newXSproto_portable("spi_fetchrow", XS__spi_fetchrow, file, "$");
        (void)newXSproto_portable("spi_prepare", XS__spi_prepare, file, "$;@");
        (void)newXSproto_portable("spi_exec_prepared", XS__spi_exec_prepared, file, "$;@");
        (void)newXSproto_portable("spi_query_prepared", XS__spi_query_prepared, file, "$;@");
        (void)newXSproto_portable("spi_freeplan", XS__spi_freeplan, file, "$");
        (void)newXSproto_portable("spi_cursor_close", XS__spi_cursor_close, file, "$");
        (void)newXSproto_portable("spi_commit", XS__spi_commit, file, "");
        (void)newXSproto_portable("spi_rollback", XS__spi_rollback, file, "");

    /* Initialisation Section */

#line 162 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.xs"
    items = 0;  /* avoid 'unused variable' warning */

#line 487 "/build/postgresql/src/build/../postgresql-18.6/src/pl/plperl/SPI.c"

    /* End of Initialisation Section */

#if PERL_VERSION_LE(5, 21, 5)
#  if PERL_VERSION_GE(5, 9, 0)
    if (PL_unitcheckav)
        call_list(PL_scopestack_ix, PL_unitcheckav);
#  endif
    XSRETURN_YES;
#else
    Perl_xs_boot_epilog(aTHX_ ax);
#endif
}

#ifdef __cplusplus
}
#endif
